"""
Callback functions and wrappers for libheif `heif_context_read_from_memory_without_copy` and `heif_context_write`.
"""

import builtins
from io import SEEK_SET
from pathlib import Path
from typing import Dict, List, Optional

from _pillow_heif_cffi import ffi, lib

from . import options
from .constants import HeifCompressionFormat
from .error import check_libheif_error
from .misc import _get_bytes, get_file_mimetype


def _get_heif_writer():
    heif_writer = ffi.new("struct heif_writer *")
    heif_writer.writer_api_version = 1
    heif_writer.write = lib.callback_write
    return heif_writer


HEIF_WRITER = _get_heif_writer()


class LibHeifCtx:
    """LibHeif reader context"""

    def __init__(self, fp, to_8bit: bool = False):
        self.to_8bit = to_8bit
        self.ctx = ffi.gc(lib.heif_context_alloc(), lib.heif_context_free)
        if hasattr(fp, "seek"):
            fp.seek(0, SEEK_SET)
        self.c_userdata = bytes(_get_bytes(fp))
        cdata_buf = ffi.from_buffer(self.c_userdata)  # For PyPy3, without this it will raise error during read.
        error = lib.heif_context_read_from_memory_without_copy(self.ctx, cdata_buf, len(cdata_buf), ffi.NULL)
        check_libheif_error(error)
        self.mimetype = get_file_mimetype(self.c_userdata)
        lib.heif_context_set_max_decoding_threads(self.ctx, options.DECODE_THREADS)

    def get_main_img_id(self) -> int:
        p_main_image_id = ffi.new("heif_item_id *")
        check_libheif_error(lib.heif_context_get_primary_image_ID(self.ctx, p_main_image_id))
        return p_main_image_id[0]

    def get_top_images_ids(self) -> List[int]:
        top_img_count = lib.heif_context_get_number_of_top_level_images(self.ctx)
        top_img_ids = ffi.new("heif_item_id[]", top_img_count)
        top_img_count = lib.heif_context_get_list_of_top_level_image_IDs(self.ctx, top_img_ids, top_img_count)
        return [top_img_ids[i] for i in range(top_img_count)]


class LibHeifCtxWrite:
    """LibHeif writer context"""

    def __init__(self, compression_format: int = HeifCompressionFormat.HEVC):
        self.ctx = ffi.gc(lib.heif_context_alloc(), lib.heif_context_free)
        p_encoder = ffi.new("struct heif_encoder **")
        error = lib.heif_context_get_encoder_for_format(self.ctx, compression_format, p_encoder)
        check_libheif_error(error)
        self.encoder = ffi.gc(p_encoder[0], lib.heif_encoder_release)
        # lib.heif_encoder_set_logging_level(self.encoder, 4)

    def set_encoder_parameters(self, enc_params: Dict[str, str], quality: Optional[int] = None):
        if quality is not None:
            if quality == -1:
                check_libheif_error(lib.heif_encoder_set_lossless(self.encoder, True))
            else:
                check_libheif_error(lib.heif_encoder_set_lossy_quality(self.encoder, quality))
        for key, value in enc_params.items():
            _value = value if isinstance(value, str) else str(value)
            check_libheif_error(
                lib.heif_encoder_set_parameter(self.encoder, key.encode("ascii"), _value.encode("ascii"))
            )

    def write(self, fp):
        __fp = self._get_fp(fp)
        c_userdata = ffi.new_handle(__fp)
        error = lib.heif_context_write(self.ctx, HEIF_WRITER, c_userdata)
        if isinstance(fp, (str, Path)):
            __fp.close()
        check_libheif_error(error)

    @staticmethod
    def _get_fp(fp):
        if isinstance(fp, (str, Path)):
            return builtins.open(fp, "wb")
        if hasattr(fp, "write"):
            return fp
        raise TypeError("`fp` must be a path to file or an object with `write` method.")


@ffi.def_extern()
def callback_write(_ctx, data, size, userdata):
    fp = ffi.from_handle(userdata)
    fp.write(ffi.buffer(data, size=size))
    return [0, 0, ffi.NULL]



import os
from pathlib import Path

from PIL import Image

import pillow_heif

os.chdir(os.path.join(os.path.dirname(os.path.dirname(os.path.abspath(__file__))), "tests"))
TARGET_FOLDER = "../converted"

# This demo converts gif to heic.
if __name__ == "__main__":
    pillow_heif.register_heif_opener()
    os.makedirs(TARGET_FOLDER, exist_ok=True)
    image_path = Path("images/non_heif/chi.gif")
    img = Image.open(image_path)
    img.save(os.path.join(TARGET_FOLDER, f"{image_path.stem}.heic"), quality=50, save_all=True)
    img.close()



import os
from pathlib import Path

from PIL import Image

import pillow_heif

os.chdir(os.path.join(os.path.dirname(os.path.dirname(os.path.abspath(__file__))), "tests"))
TARGET_FOLDER = "../converted"

# Create HEIF file containing `RGB_16.png` and `RGBA_16.png`
if __name__ == "__main__":
    os.makedirs(TARGET_FOLDER, exist_ok=True)
    image1_path = Path("images/non_heif/RGB_16.png")
    image2_path = Path("images/non_heif/RGBA_16.png")
    img1 = Image.open(image1_path)
    img2 = Image.open(image2_path)
    result_path = os.path.join(TARGET_FOLDER, "RGB_A_8.heic")
    heif_file = pillow_heif.from_pillow(img1)
    heif_file.add_from_pillow(img2)  # it can be replaced with `append_images` parameter.
    heif_file.save(result_path, quality=35)



from io import BytesIO

import pytest
from helpers import assert_image_similar, hevc_enc
from PIL import Image, ImageOps

import pillow_heif

pillow_heif.register_heif_opener()


def get_xmp_with_orientation(orientation: int, style=1) -> str:
    xmp_1 = (
        '<?xpacket begin="\xef\xbb\xbf" id="W5M0MpCehiHzreSzNTczkc9d"?>\n'
        '<x:xmpmeta xmlns:x="adobe:ns:meta/" x:xmptk="XMP Core 5.4.0">\n'
        ' <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#">\n'
        '  <rdf:Description rdf:about=""\n'
        '    xmlns:exif="http://ns.adobe.com/exif/1.0/"\n'
        '    xmlns:tiff="http://ns.adobe.com/tiff/1.0/"\n'
        '   tiff:Orientation="'
        + str(orientation)
        + '"/>\n </rdf:RDF>\n</x:xmpmeta>\n<?xpacket end="r"?>'
    )
    xmp_2 = (
        '<?xpacket begin="\xef\xbb\xbf" id="W5M0MpCehiHzreSzNTczkc9d"?>\n'
        '<x:xmpmeta xmlns:x="adobe:ns:meta/" x:xmptk="Image::ExifTool 12.30">\n'
        '<rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#">\n\n'
        ' <rdf:Description rdf:about=""\n'
        '  xmlns:exif="http://ns.adobe.com/exif/1.0/">\n'
        "  <exif:ColorSpace>1</exif:ColorSpace>\n"
        "  <exif:PixelXDimension>384</exif:PixelXDimension>\n"
        "  <exif:PixelYDimension>512</exif:PixelYDimension>\n"
        " </rdf:Description>\n\n"
        ' <rdf:Description rdf:about=""\n'
        '  xmlns:tiff="http://ns.adobe.com/tiff/1.0/">\n'
        "  <tiff:Orientation>"
        + str(orientation)
        + '</tiff:Orientation>\n </rdf:Description>\n</rdf:RDF>\n</x:xmpmeta>\n<?xpacket end="r"?>'
    )
    return xmp_1 if style == 1 else xmp_2


@pytest.mark.skipif(not hevc_enc(), reason="Requires HEVC encoder.")
@pytest.mark.parametrize("orientation", (1, 2, 3, 4, 5, 6, 7, 8))
@pytest.mark.parametrize("im_format", ("JPEG", "PNG"))
def test_exif_orientation(orientation, im_format):
    out_im = BytesIO()
    out_heif_im = BytesIO()
    im = Image.effect_mandelbrot((256, 128), (-3, -2.5, 2, 2.5), 100).crop((0, 0, 256, 96))
    im = im.convert(mode="RGB")
    exif_data = Image.Exif()
    exif_data[0x0112] = orientation
    im.save(out_im, format=im_format, exif=exif_data.tobytes())
    im = Image.open(out_im)
    assert im.getexif()[0x0112] == orientation
    # Saving image with EXIF to HEIF
    im.save(out_heif_im, format="HEIF", quality=-1)
    im_heif = Image.open(out_heif_im)
    im_heif_exif = im_heif.getexif()
    assert 0x0112 not in im_heif_exif or im_heif_exif[0x0112] == 1
    transposed_im = ImageOps.exif_transpose(im)
    assert_image_similar(transposed_im, im_heif)
    if orientation > 1:
        with pytest.raises(AssertionError):
            assert_image_similar(im, im_heif)


@pytest.mark.parametrize("orientation", (1, 2, 3, 4, 5, 6, 7, 8))
def test_png_xmp_orientation(orientation):
    im = Image.effect_mandelbrot((256, 128), (-3, -2.5, 2, 2.5), 100).crop((0, 0, 256, 96))
    im = im.convert(mode="RGB")
    xmp = get_xmp_with_orientation(orientation)
    im.info["XML:com.adobe.xmp"] = xmp
    # We are testing next two lines. They are equal to `save+open` operations.
    im_heif = pillow_heif.from_pillow(im)
    im_heif = im_heif[0].to_pillow()
    assert im_heif.info["xmp"]
    assert isinstance(im_heif.info["xmp"], bytes)
    transposed_im = ImageOps.exif_transpose(im)
    assert_image_similar(transposed_im, im_heif)
    if orientation > 1:
        with pytest.raises(AssertionError):
            assert_image_similar(im, im_heif)


@pytest.mark.skipif(not hevc_enc(), reason="Requires HEVC encoder.")
@pytest.mark.parametrize("orientation", (1, 2, 3, 4, 5, 6, 7, 8))
def test_heif_exif_orientation(orientation):
    im = Image.effect_mandelbrot((256, 128), (-3, -2.5, 2, 2.5), 100).crop((0, 0, 256, 96))
    im = im.convert(mode="RGB")
    out_im_heif = BytesIO()
    exif_data = Image.Exif()
    exif_data[0x0112] = orientation
    im.save(out_im_heif, format="HEIF", exif=exif_data.tobytes(), quality=-1)
    im_heif = Image.open(out_im_heif)
    # We should ignore all EXIF rotation flags for HEIF
    if orientation > 1:
        assert im_heif.info["original_orientation"] == orientation
    else:
        assert im_heif.info.get("original_orientation", None) is None
    im_heif_exif = im_heif.getexif()
    assert 0x0112 not in im_heif_exif or im_heif_exif[0x0112] == 1
    assert_image_similar(im, im_heif)


@pytest.mark.skipif(not hevc_enc(), reason="Requires HEVC encoder.")
@pytest.mark.parametrize("orientation", (1, 2, 3, 4, 5, 6, 7, 8))
def test_heif_xmp_orientation(orientation):
    im = Image.effect_mandelbrot((256, 128), (-3, -2.5, 2, 2.5), 100).crop((0, 0, 256, 96))
    im = im.convert(mode="RGB")
    xmp = get_xmp_with_orientation(orientation)
    out_im_heif = BytesIO()
    im.save(out_im_heif, format="HEIF", xmp=xmp.encode("utf-8"), quality=-1)
    im_heif = Image.open(out_im_heif)
    # We should ignore all XMP rotation flags for HEIFss
    if orientation > 1:
        assert im_heif.info["original_orientation"] == orientation
    else:
        assert im_heif.info.get("original_orientation", None) is None
    assert_image_similar(im, im_heif)


@pytest.mark.skipif(not hevc_enc(), reason="Requires HEVC encoder.")
@pytest.mark.parametrize("orientation", (1, 2))
def test_heif_xmp_orientation_exiftool(orientation):
    im = Image.effect_mandelbrot((256, 128), (-3, -2.5, 2, 2.5), 100).crop((0, 0, 256, 96))
    im = im.convert(mode="RGB")
    xmp = get_xmp_with_orientation(orientation, style=2)
    out_im_heif = BytesIO()
    im.save(out_im_heif, format="HEIF", xmp=xmp.encode("utf-8"), quality=-1)
    im_heif = Image.open(out_im_heif)
    # We should ignore all XMP rotation flags for HEIFss
    if orientation > 1:
        assert im_heif.info["original_orientation"] == orientation
    else:
        assert im_heif.info.get("original_orientation", None) is None
    assert_image_similar(im, im_heif)


@pytest.mark.skipif(not hevc_enc(), reason="Requires HEVC encoder.")
@pytest.mark.parametrize("orientation", (1, 2))
def test_heif_xmp_orientation_with_exif_eq_1(orientation):
    im = Image.effect_mandelbrot((256, 128), (-3, -2.5, 2, 2.5), 100).crop((0, 0, 256, 96))
    im = im.convert(mode="RGB")
    xmp = get_xmp_with_orientation(orientation)
    out_im_heif = BytesIO()
    exif_data = Image.Exif()
    exif_data[0x0112] = 1
    im.save(out_im_heif, format="HEIF", exif=exif_data.tobytes(), xmp=xmp.encode("utf-8"), quality=-1)
    im_heif = Image.open(out_im_heif)
    # We should ignore all XMP rotation flags for HEIF
    if orientation > 1:
        assert im_heif.info["original_orientation"] == orientation
    else:
        assert im_heif.info.get("original_orientation", None) is None
    assert_image_similar(im, im_heif)


@pytest.mark.skipif(not hevc_enc(), reason="Requires HEVC encoder.")
@pytest.mark.parametrize("orientation", (1, 2))
@pytest.mark.parametrize("im_format", ("JPEG", "PNG"))
def test_exif_heif_exif_orientation(orientation, im_format):
    out_im = BytesIO()
    out_heif_im = BytesIO()
    im = Image.effect_mandelbrot((256, 128), (-3, -2.5, 2, 2.5), 100).crop((0, 0, 256, 96))
    im = im.convert(mode="RGB")
    exif_data = Image.Exif()
    exif_data[0x0112] = orientation
    im.save(out_im, format=im_format, exif=exif_data.tobytes())
    im = Image.open(out_im)
    assert im.getexif()[0x0112] == orientation
    # Saving image with EXIF to HEIF
    im.save(out_heif_im, format="HEIF", quality=-1, exif=exif_data.tobytes())
    im_heif = Image.open(out_heif_im)
    im_heif_exif = im_heif.getexif()
    assert 0x0112 not in im_heif_exif or im_heif_exif[0x0112] == 1
    transposed_im = ImageOps.exif_transpose(im)
    assert_image_similar(transposed_im, im_heif)
    if orientation > 1:
        with pytest.raises(AssertionError):
            assert_image_similar(im, im_heif)



import os
from pathlib import Path

import pillow_heif

os.chdir(os.path.join(os.path.dirname(os.path.dirname(os.path.abspath(__file__))), "tests"))
TARGET_FOLDER = "../converted"

if __name__ == "__main__":
    os.makedirs(TARGET_FOLDER, exist_ok=True)
    image_path = Path("images/heif_other/nokia/alpha.heic")
    heif_file = pillow_heif.open_heif(image_path)
    result_path = os.path.join(TARGET_FOLDER, f"{image_path.stem}.heic")
    pillow_heif.add_thumbnails(heif_file, [256, 372])
    heif_file.save(result_path, quality=35)

    # Next code will remove thumbnails
    heif_file = pillow_heif.open_heif(result_path)
    del heif_file.thumbnails[0]  # remove first thumbnail
    del heif_file.thumbnails[0]  # remove second thumbnail
    result_path = os.path.join(TARGET_FOLDER, f"{image_path.stem}_no_thumbs.heic")
    heif_file.save(result_path, quality=35)



"""
Different miscellaneous helper functions.

Mostly for internal use, so prototypes can change between versions.
"""

import builtins
import pathlib
import re
from struct import pack, unpack
from typing import Union


def set_orientation(info: dict, orientation: int = 1) -> Union[int, None]:
    """Sets orientation in ``EXIF`` to ``1`` by default if any orientation present.
    Removes ``XMP`` orientation tag if it is present.
    In Pillow plugin mode it called automatically for images.
    When ``pillow_heif`` used in ``standalone`` mode, if you wish you can call it manually.

    .. note:: If there is no orientation tag, this function will not add it and do nothing.

        If both XMP and EXIF orientation tags present, EXIF orientation tag will be returned,
        but both tags will be removed.

    :param info: `info` dictionary from `~PIL.Image.Image` or `~pillow_heif.HeifImage`.
    :param orientation: int value of EXIF or XMP orientation tag.
    :returns: Original orientation or None if it is absent."""

    original_orientation = None
    if info.get("exif", None):
        try:
            tif_tag = info["exif"]
            if tif_tag.startswith(b"Exif\x00\x00"):
                tif_tag = tif_tag[6:]
            endian_mark = "<" if tif_tag[0:2] == b"\x49\x49" else ">"
            pointer = unpack(endian_mark + "L", tif_tag[4:8])[0]
            tag_count = unpack(endian_mark + "H", tif_tag[pointer : pointer + 2])[0]
            offset = pointer + 2
            for tag_n in range(tag_count):
                pointer = offset + 12 * tag_n
                if unpack(endian_mark + "H", tif_tag[pointer : pointer + 2])[0] != 274:
                    continue
                value = tif_tag[pointer + 8 : pointer + 12]
                _original_orientation = unpack(endian_mark + "H", value[0:2])[0]
                if _original_orientation != 1:
                    original_orientation = _original_orientation
                    p_value = 6 + pointer + 8
                    new_orientation = pack(endian_mark + "H", orientation)
                    info["exif"] = info["exif"][:p_value] + new_orientation + info["exif"][p_value + 2 :]
                    break
        except Exception:  # noqa # pylint: disable=broad-except
            pass
    if info.get("xmp", None):
        xmp_data = info["xmp"].decode("utf-8")
        match = re.search(r'tiff:Orientation(="|>)([0-9])', xmp_data)
        if match:
            if original_orientation is None and int(match[2]) != 1:
                original_orientation = int(match[2])
            xmp_data = re.sub(r'tiff:Orientation="([0-9])"', "", xmp_data)
            xmp_data = re.sub(r"<tiff:Orientation>([0-9])</tiff:Orientation>", "", xmp_data)
            info["xmp"] = xmp_data.encode("utf-8")
    return original_orientation


def get_file_mimetype(fp) -> str:
    """Gets the MIME type of the HEIF(or AVIF) object.`

    :param fp: A filename (string), pathlib.Path object, file object or bytes.
        The file object must implement ``file.read``, ``file.seek`` and ``file.tell`` methods,
        and be opened in binary mode.
    :returns: "image/heic", "image/heif", "image/heic-sequence", "image/heif-sequence",
        "image/avif", "image/avif-sequence" or "".
    """

    heif_brand = _get_bytes(fp, 12)[8:]
    if heif_brand:
        if heif_brand == b"avif":
            return "image/avif"
        if heif_brand == b"avis":
            return "image/avif-sequence"
        if heif_brand in (b"heic", b"heix", b"heim", b"heis"):
            return "image/heic"
        if heif_brand in (b"hevc", b"hevx", b"hevm", b"hevs"):
            return "image/heic-sequence"
        if heif_brand == b"mif1":
            return "image/heif"
        if heif_brand == b"msf1":
            return "image/heif-sequence"
    return ""


def _get_bytes(fp, length=None) -> bytes:
    if isinstance(fp, (str, pathlib.Path)):
        with builtins.open(fp, "rb") as file:
            return file.read(length or -1)
    if hasattr(fp, "read"):
        offset = fp.tell() if hasattr(fp, "tell") else None
        result = fp.read(length or -1)
        if offset is not None and hasattr(fp, "seek"):
            fp.seek(offset)
        return result
    return bytes(fp)[:length]



import sys
from os import getenv, path
from pathlib import Path
from re import sub
from shutil import copy
from subprocess import DEVNULL, PIPE, run
from warnings import warn

from cffi import FFI

from libheif import linux_build_libs

ffi = FFI()
with open("libheif/public_api.h", "r", encoding="utf-8") as f:
    libheif_definitions = f.read()
    if getenv("READTHEDOCS", "False") == "True":
        # As ReadTheDocs has pre-installed libheif.so and we do not have root privileges to uninstall it
        # remove lines from `public_api.h` with `REMOVE_FOR_RTD` string.
        libheif_definitions = sub(r".*REMOVE_FOR_RTD\n?", "", libheif_definitions)
    ffi.cdef(libheif_definitions)

ffi.cdef(
    """
    extern "Python" struct heif_error callback_write(struct heif_context*, const void*, size_t, void*);
"""
)

with open("pillow_heif/helpers.h", "r", encoding="utf-8") as f:
    ffi.cdef(f.read())


include_dirs = ["/usr/local/include", "/usr/include"]
library_dirs = ["/usr/local/lib", "/usr/lib64", "/usr/lib", "/lib"]

include_path_prefix = ""
insert = False
if sys.platform.lower() == "darwin":
    include_path_prefix = getenv("HOMEBREW_PREFIX")
    if not include_path_prefix:
        _result = run(["brew", "--prefix"], stderr=DEVNULL, stdout=PIPE, check=False)
        if not _result.returncode and _result.stdout is not None:
            include_path_prefix = _result.stdout.decode("utf-8").rstrip("\n")
    if not include_path_prefix:
        include_path_prefix = "/opt/local"
elif sys.platform.lower() == "win32":
    include_path_prefix = getenv("MSYS2_PREFIX")
    if include_path_prefix is None:
        include_path_prefix = "C:\\msys64\\mingw64"
        warn(f"MSYS2_PREFIX environment variable is not set. Assuming `MSYS2_PREFIX={include_path_prefix}`")
else:
    include_path_prefix = linux_build_libs.build_libs()

# Need to include "lib" directory to find "heif" library.
include_path_prefix_lib = path.join(include_path_prefix, "lib")
if include_path_prefix_lib not in library_dirs:
    library_dirs.append(include_path_prefix_lib)

# MSYS2: rename "libheif.dll.a" to "libheif.lib"
if sys.platform.lower() == "win32":
    lib_export_file = Path(path.join(include_path_prefix_lib, "libheif.dll.a"))
    if lib_export_file.is_file():
        copy(lib_export_file, path.join(include_path_prefix_lib, "libheif.lib"))
    else:
        warn("If you build this with MSYS2, you should not see this warning.")

# Adds project root to `include` path
include_dirs.append(path.dirname(path.dirname(path.abspath(__file__))))

ffi.set_source(
    "_pillow_heif_cffi",
    r"""
    #include "libheif/public_api.h"
    #include "pillow_heif/helpers.c"
    """,
    include_dirs=include_dirs,
    library_dirs=library_dirs,
    libraries=["libheif"] if sys.platform.lower() == "win32" else ["heif"],
    extra_compile_args=["/d2FH4-"] if sys.platform.lower() == "win32" else [],
)

if __name__ == "__main__":
    ffi.compile(verbose=True)



"""
Functions to get version and encoders/decoders info of embedded C libraries.
"""

from _pillow_heif_cffi import ffi, lib

from .constants import HeifCompressionFormat


def libheif_version() -> str:
    """Wrapper around `libheif.heif_get_version`"""
    return ffi.string(lib.heif_get_version()).decode()


def have_decoder_for_format(format_id: HeifCompressionFormat) -> bool:
    """Wrapper around `libheif.heif_have_decoder_for_format`"""
    return lib.heif_have_decoder_for_format(format_id)


def have_encoder_for_format(format_id: HeifCompressionFormat) -> bool:
    """Wrapper around `libheif.heif_have_encoder_for_format`"""
    return lib.heif_have_encoder_for_format(format_id)


def libheif_info() -> dict:
    """Returns dictionary with available decoders & encoders and libheif version.
    Keys are `versions`, `decoders`, `encoders`.

    {'version': {
        'libheif': '1.14.0',
        'x265': 'x265 HEVC encoder (3.4+31-6722fce1f)',
        'aom': 'AOMedia Project AV1 Encoder 3.5.0'
        },
     'decoders': {'HEVC': 1, 'AV1': 1, 'AVC': 0},
     'encoders': {'HEVC': 1, 'AV1': 1, 'AVC': 0}
    }"""

    decoders = {}
    encoders = {}
    for format_id in (HeifCompressionFormat.HEVC, HeifCompressionFormat.AV1, HeifCompressionFormat.AVC):
        decoders[format_id.name] = have_decoder_for_format(format_id)
        encoders[format_id.name] = have_encoder_for_format(format_id)

    _version = {"libheif": libheif_version(), "x265": "", "aom": ""}
    p_enc_desc = ffi.new("struct heif_encoder_descriptor**")
    if lib.heif_context_get_encoder_descriptors(ffi.NULL, HeifCompressionFormat.HEVC, ffi.NULL, p_enc_desc, 1):
        p_enc_name = lib.heif_encoder_descriptor_get_name(p_enc_desc[0])
        _version["x265"] = ffi.string(p_enc_name).decode()
    if lib.heif_context_get_encoder_descriptors(ffi.NULL, HeifCompressionFormat.AV1, ffi.NULL, p_enc_desc, 1):
        p_enc_name = lib.heif_encoder_descriptor_get_name(p_enc_desc[0])
        _version["aom"] = ffi.string(p_enc_name).decode()

    return {"version": _version, "decoders": decoders, "encoders": encoders}



import os
from io import SEEK_END, BytesIO
from time import perf_counter

import pytest
from helpers import aom_enc, create_heif, hevc_enc
from PIL import Image

from pillow_heif import (
    from_pillow,
    open_heif,
    options,
    register_avif_opener,
    register_heif_opener,
)

os.chdir(os.path.dirname(os.path.abspath(__file__)))


@pytest.mark.skipif(not hevc_enc(), reason="No HEVC encoder.")
@pytest.mark.skipif(not aom_enc(), reason="No AVIF encoder.")
@pytest.mark.parametrize("register_opener", (register_avif_opener, register_heif_opener))
def test_options_change_from_plugin_registering(register_opener):
    try:
        register_opener(thumbnails=False, quality=69, save_to_12bit=True, decode_threads=3)
        assert not options.THUMBNAILS
        assert options.QUALITY == 69
        assert options.SAVE_HDR_TO_12_BIT
        assert options.DECODE_THREADS == 3
    finally:
        options.THUMBNAILS = True
        options.QUALITY = None
        options.SAVE_HDR_TO_12_BIT = False
        options.DECODE_THREADS = 4


@pytest.mark.skipif(not hevc_enc(), reason="No HEVC encoder.")
def test_thumbnails_option():
    heif_buf = create_heif((128, 128), [64])
    try:
        heif_file = open_heif(heif_buf)
        assert heif_file.thumbnails
        assert str(heif_file.thumbnails[0]).find("bytes image data") == -1
        assert heif_file.thumbnails[0].data
        assert str(heif_file.thumbnails[0]).find("bytes image data") != -1
        # disabling thumbnails and checking them not to be present
        options.THUMBNAILS = False
        heif_file = open_heif(heif_buf)
        assert not heif_file.thumbnails
    finally:
        options.THUMBNAILS = True


@pytest.mark.skipif(not hevc_enc(), reason="No HEVC encoder.")
@pytest.mark.skipif(not aom_enc(), reason="No AVIF encoder.")
@pytest.mark.parametrize("save_format", ("HEIF", "AVIF"))
def test_quality_option(save_format):
    try:
        image = from_pillow(Image.linear_gradient(mode="L"))
        options.QUALITY = 10
        out_buf_1_q10 = BytesIO()
        image.save(out_buf_1_q10, format=save_format)
        options.QUALITY = 30
        # passing manual `quality` has higher priority then one in `options`
        out_buf_2_q10 = BytesIO()
        image.save(out_buf_2_q10, quality=10, format=save_format)
        assert out_buf_1_q10.seek(0, SEEK_END) == out_buf_2_q10.seek(0, SEEK_END)
        # image with quality 30% must have bigger size then 10%
        out_buf_3_q30 = BytesIO()
        image.save(out_buf_3_q30, format=save_format)
        assert out_buf_1_q10.seek(0, SEEK_END) < out_buf_3_q30.seek(0, SEEK_END)
    finally:
        options.QUALITY = None


@pytest.mark.skipif(os.cpu_count() < 2, reason="Requires at least a processor with two cores.")
@pytest.mark.skipif(os.getenv("TEST_DECODE_THREADS", "1") == "0", reason="TEST_DECODE_THREADS set to `0`")
def test_decode_threads():
    test_image = "images/heif_other/arrow.heic"  # not all images can be decoded using more than one thread
    # As we do not know real performance of hardware, measure relative
    try:
        options.DECODE_THREADS = 1
        start_time_one_thread = perf_counter()
        open_heif(test_image, convert_hdr_to_8bit=False).load()
        total_time_one_thread = perf_counter() - start_time_one_thread
        options.DECODE_THREADS = os.cpu_count()
        start_time_multiply_threads = perf_counter()
        open_heif(test_image, convert_hdr_to_8bit=False).load()
        total_time_multiply_threads = perf_counter() - start_time_multiply_threads
        # decoding in multiply threads should be faster at least by 15%
        assert total_time_one_thread > total_time_multiply_threads * 1.15
    finally:
        options.DECODE_THREADS = 4


def test_plugin_register_unknown_option():
    with pytest.warns(UserWarning, match="Unknown option: unknown_option"):
        register_heif_opener(unknown_option=12345)



"""
Custom exception that can be raised during library calls.
"""

from _pillow_heif_cffi import ffi

from .constants import HeifErrorCode


class HeifError(ValueError):
    """
    Raised in :py:meth:`pillow_heif.open_heif` if an image cannot be opened or corrupted.

    Also, can be raised during image decoding or during saving.
    """

    def __init__(self, *, code, subcode, message):
        super().__init__(code, subcode, message)
        self.code = code
        self.subcode = subcode
        self.message = message

    def __str__(self):
        return f"Code: {self.code}, Subcode: {self.subcode}, Message: `{self.message}`"

    def __repr__(self):
        return f"{repr(HeifErrorCode(self.code))}, {self.subcode}, {self.message}"


def check_libheif_error(error_struct) -> None:
    """
    Helper function. Checks returned result error_struct from libheif calls and raise exception if error.

    :param error_struct: ``heif_error`` struct from libheif.

    :exception HeifError: If there is an error code.
    """

    if error_struct.code == HeifErrorCode.OK:
        return
    raise HeifError(
        code=error_struct.code,
        subcode=error_struct.subcode,
        message=ffi.string(error_struct.message).decode(),
    )



from io import BytesIO

import pytest
from helpers import aom_enc, create_heif, hevc_enc
from PIL import Image

import pillow_heif

pillow_heif.register_avif_opener()
pillow_heif.register_heif_opener()


@pytest.mark.skipif(not aom_enc(), reason="Requires AVIF encoder.")
@pytest.mark.skipif(not hevc_enc(), reason="Requires HEVC encoder.")
@pytest.mark.parametrize("save_format", ("HEIF", "AVIF"))
def test_heif_primary_image(save_format):
    heif_buf = create_heif((64, 64), n_images=3, primary_index=1, format=save_format)
    heif_file = pillow_heif.open_heif(heif_buf)
    assert heif_file.primary_index() == 1
    assert heif_file[1].info["primary"]
    out_buf = BytesIO()
    heif_file.save(out_buf, quality=1, format=save_format)
    heif_file_out = pillow_heif.open_heif(out_buf)
    assert heif_file_out.primary_index() == 1
    assert heif_file_out[1].info["primary"]
    heif_file.save(out_buf, quality=1, primary_index=0, format=save_format)
    assert heif_file.primary_index() == 1
    heif_file_out = pillow_heif.open_heif(out_buf)
    assert heif_file_out.primary_index() == 0
    heif_file.save(out_buf, quality=1, primary_index=-1, format=save_format)
    assert heif_file.primary_index() == 1
    heif_file_out = pillow_heif.open_heif(out_buf)
    assert heif_file_out.primary_index() == 2
    heif_file.save(out_buf, quality=1, primary_index=99, format=save_format)
    assert heif_file.primary_index() == 1
    heif_file_out = pillow_heif.open_heif(out_buf)
    assert heif_file_out.primary_index() == 2


@pytest.mark.skipif(not aom_enc(), reason="Requires AVIF encoder.")
@pytest.mark.skipif(not hevc_enc(), reason="Requires HEVC encoder.")
@pytest.mark.parametrize("save_format", ("HEIF", "AVIF"))
def test_pillow_primary_image(save_format):
    heif_buf = create_heif((64, 64), n_images=3, primary_index=1, format=save_format)
    heif_file = Image.open(heif_buf)
    assert heif_file.tell() == 1
    assert heif_file.info["primary"]
    out_buf = BytesIO()
    heif_file.save(out_buf, format=save_format, save_all=True, quality=1)
    heif_file_out = Image.open(out_buf)
    assert heif_file_out.tell() == 1
    assert heif_file_out.info["primary"]
    heif_file.save(out_buf, format=save_format, save_all=True, quality=1, primary_index=0)
    assert heif_file.tell() == 1
    heif_file_out = Image.open(out_buf)
    assert heif_file_out.tell() == 0
    heif_file.save(out_buf, format=save_format, save_all=True, quality=1, primary_index=-1)
    assert heif_file.tell() == 1
    heif_file_out = Image.open(out_buf)
    assert heif_file_out.tell() == 2
    heif_file.save(out_buf, format=save_format, save_all=True, quality=1, primary_index=99)
    assert heif_file.tell() == 1
    heif_file_out = Image.open(out_buf)
    assert heif_file_out.tell() == 2


@pytest.mark.skipif(not aom_enc(), reason="Requires AVIF encoder.")
@pytest.mark.skipif(not hevc_enc(), reason="Requires HEVC encoder.")
@pytest.mark.parametrize("save_format", ("HEIF", "AVIF"))
def test_heif_info_changing(save_format):
    xmp = b"LeagueOf"
    exif_desc_value = "this is a desc"
    exif = Image.Exif()
    exif[0x010E] = exif_desc_value
    heif_buf = create_heif((64, 64), n_images=3, primary_index=1, exif=exif.tobytes(), xmp=xmp, format=save_format)
    im = pillow_heif.open_heif(heif_buf)
    assert im.info["primary"] and len(im.info["exif"]) > 0 and im.info["xmp"]
    out_buf = BytesIO()
    # Remove exif and xmp of a Primary Image.
    im.save(out_buf, exif=None, xmp=None, format=save_format)
    assert im.info["primary"] and len(im.info["exif"]) > 0 and im.info["xmp"]
    im_out = pillow_heif.open_heif(out_buf)
    for i in range(3):
        if i == 1:
            assert im_out[i].info["primary"] and not im_out[i].info["exif"] and not im_out[i].info["xmp"]
        else:
            assert not im_out[i].info["primary"] and not im_out[i].info["exif"] and not im_out[i].info["xmp"]
    # Set exif and xmp of all images. Change Primary Image to be last.
    for i in range(3):
        im[i].info["xmp"] = xmp
        im[i].info["exif"] = exif.tobytes()
        im[i].info["primary"] = False if i != 2 else True
    im.save(out_buf, format=save_format)
    im_out = pillow_heif.open_heif(out_buf)
    assert im_out.info["primary"]
    assert im_out.primary_index() == 2
    for i in range(3):
        assert im_out[i].info["exif"] and im_out[i].info["xmp"]
    # Remove `primary`, `xmp`, `exif` from info dict.
    for i in range(3):
        im[i].info.pop("xmp")
        im[i].info.pop("exif")
        im[i].info.pop("primary")
    im.save(out_buf, format=save_format)
    im_out = pillow_heif.open_heif(out_buf)
    assert im_out.info["primary"]
    assert im_out.primary_index() == 0
    for i in range(3):
        assert not im_out[i].info["exif"] and not im_out[i].info["xmp"]


@pytest.mark.skipif(not aom_enc(), reason="Requires AVIF encoder.")
@pytest.mark.skipif(not hevc_enc(), reason="Requires HEVC encoder.")
@pytest.mark.parametrize("save_format", ("HEIF", "AVIF"))
def test_pillow_info_changing(save_format):
    xmp = b"LeagueOf"
    exif_desc_value = "this is a desc"
    exif = Image.Exif()
    exif[0x010E] = exif_desc_value
    heif_buf = create_heif((64, 64), n_images=3, primary_index=1, exif=exif.tobytes(), xmp=xmp, format=save_format)
    im = Image.open(heif_buf)
    assert im.info["primary"] and len(im.info["exif"]) > 0 and im.info["xmp"]
    out_buf = BytesIO()
    # Remove exif and xmp of a Primary Image.
    im.save(out_buf, format=save_format, save_all=True, exif=None, xmp=None)
    assert im.info["primary"] and len(im.info["exif"]) > 0 and im.info["xmp"]
    im_out = Image.open(out_buf)
    for i in range(3):
        im_out.seek(i)
        if i == 1:
            assert im_out.info["primary"] and not im_out.info["exif"] and not im_out.info["xmp"]
        else:
            assert not im_out.info["primary"] and not im_out.info["exif"] and not im_out.info["xmp"]
    # Set exif and xmp of all images. Change Primary Image to be last.
    for i in range(3):
        im.seek(i)
        im.info["xmp"] = xmp
        im.info["exif"] = exif.tobytes()
        im.info["primary"] = False if i != 2 else True
    im.save(out_buf, format=save_format, save_all=True)
    im_out = Image.open(out_buf)
    assert im_out.info["primary"]
    assert im_out.tell() == 2
    for i in range(3):
        im_out.seek(i)
        assert im_out.info["exif"] and im_out.info["xmp"]
    # Remove `primary`, `xmp`, `exif` from info dict.
    for i in range(3):
        im.seek(i)
        im.info.pop("xmp")
        im.info.pop("exif")
        im.info.pop("primary")
    im.save(out_buf, format=save_format, save_all=True)
    im_out = Image.open(out_buf)
    assert im_out.info["primary"]
    assert im_out.tell() == 0
    for i in range(3):
        im_out.seek(i)
        assert not im_out.info["exif"] and not im_out.info["xmp"]


@pytest.mark.skipif(not aom_enc(), reason="Requires AVIF encoder.")
@pytest.mark.skipif(not hevc_enc(), reason="Requires HEVC encoder.")
@pytest.mark.parametrize("save_format", ("HEIF", "AVIF"))
def test_heif_iptc_metadata(save_format):
    heif_buf = create_heif((64, 64), format=save_format)
    data = (
        b"\x1c\x02Z\x00\x08Budapest"
        b"\x1c\x02e\x00\x07Hungary"
        b"\x1c\x02\x19\x00\x03HvK"
        b"\x1c\x02\x19\x00\x042006"
        b"\x1c\x02\x19\x00\x06summer"
        b"\x1c\x02\x19\x00\x04July"
        b"\x1c\x02\x19\x00\x07holiday"
        b"\x1c\x02\x19\x00\x07Hungary"
        b"\x1c\x02\x19\x00\x08Budapest"
    )
    iptc_metadata = {"type": "iptc", "data": data, "content_type": b""}
    im = pillow_heif.open_heif(heif_buf)
    im.info["metadata"].append(iptc_metadata)
    out_buf = BytesIO()
    im.save(out_buf, format=save_format)
    im_out = pillow_heif.open_heif(out_buf)
    assert im_out.info["metadata"]
    assert im_out.info["metadata"][0]["type"] == "iptc"
    assert im_out.info["metadata"][0]["data"] == data


@pytest.mark.skipif(not aom_enc(), reason="Requires AVIF encoder.")
@pytest.mark.skipif(not hevc_enc(), reason="Requires HEVC encoder.")
@pytest.mark.parametrize("save_format", ("HEIF", "AVIF"))
def test_pillow_iptc_metadata(save_format):
    heif_buf = create_heif((64, 64), format=save_format)
    data = (
        b"\x1c\x02Z\x00\x08Budapest"
        b"\x1c\x02e\x00\x07Hungary"
        b"\x1c\x02\x19\x00\x03HvK"
        b"\x1c\x02\x19\x00\x042006"
        b"\x1c\x02\x19\x00\x06summer"
        b"\x1c\x02\x19\x00\x04July"
        b"\x1c\x02\x19\x00\x07holiday"
        b"\x1c\x02\x19\x00\x07Hungary"
        b"\x1c\x02\x19\x00\x08Budapest"
    )
    iptc_metadata = {"type": "iptc", "data": data, "content_type": b""}
    im = Image.open(heif_buf)
    im.info["metadata"].append(iptc_metadata)
    out_buf = BytesIO()
    im.save(out_buf, format=save_format)
    im_out = Image.open(out_buf)
    assert im_out.info["metadata"]
    assert im_out.info["metadata"][0]["type"] == "iptc"
    assert im_out.info["metadata"][0]["data"] == data



import gc
import sys
from io import BytesIO
from os import chdir, path
from pathlib import Path
from platform import machine

import helpers
import pytest
from PIL import Image

import pillow_heif

pytest.importorskip("pympler", reason="`pympler` not installed")
pytest.importorskip("numpy", reason="`numpy` not installed")

chdir(path.join(path.dirname(path.dirname(path.abspath(__file__))), "tests"))
pillow_heif.register_heif_opener()


def perform_open_save(iterations, image_path):
    for _ in range(iterations):
        image = Image.open(image_path)
        assert getattr(image, "heif_file") is not None
        if helpers.hevc_enc():
            out_buf = BytesIO()
            image.save(out_buf, quality=20, format="HEIF", save_all=True)
        elif helpers.aom_enc():
            out_buf = BytesIO()
            image.save(out_buf, quality=20, format="AVIF", save_all=True)


@pytest.mark.skipif(sys.executable.lower().find("pypy") != -1, reason="Disabled on PyPy.")
@pytest.mark.skipif(not helpers.hevc_enc(), reason="Requires HEVC encoder.")
def test_open_save_objects_leaks():
    from pympler import summary, tracker

    image_path = Path("images/heif/zPug_3.heic")
    perform_open_save(1, image_path)
    gc.collect()
    _summary1 = tracker.SummaryTracker().create_summary()
    _summary1 = tracker.SummaryTracker().create_summary()  # noqa
    gc.collect()
    gc.set_debug(gc.DEBUG_SAVEALL)
    perform_open_save(5, image_path)
    gc.collect()
    summary2 = tracker.SummaryTracker().create_summary()
    results = summary._sweep(summary.get_diff(_summary1, summary2))  # noqa
    summary.print_(results)
    for result in results:
        # look for strings like: `_cffi_backend.__CDataGCP`, `_cffi_backend.__CDataOwnGC`
        assert result[0].find("cffi") == -1
        # look for strings like: `pillow_heif._libheif_ctx.LibHeifCtx`
        assert result[0].find("pillow_heif") == -1


def _get_mem_usage():
    from resource import RUSAGE_SELF, getrusage

    if sys.platform == "darwin":
        return int(getrusage(RUSAGE_SELF).ru_maxrss / 1024)  # Kb
    return getrusage(RUSAGE_SELF).ru_maxrss  # Kb


@pytest.mark.skipif(sys.platform.lower() == "win32", reason="requires Unix or macOS")
@pytest.mark.skipif(machine().find("x86_64") == -1, reason="run only on x86_64")
def test_open_to_numpy_mem_leaks():
    import numpy as np

    mem_limit = None
    im_path = Path("images/heif/L_10.heif")
    for i in range(500):
        heif_file = pillow_heif.open_heif(im_path, convert_hdr_to_8bit=False)
        _array = np.asarray(heif_file[0])  # noqa
        _array = None  # noqa
        gc.collect()
        mem = _get_mem_usage()
        if i < 300:
            mem_limit = mem + 1
            continue
        assert mem <= mem_limit, f"memory usage limit exceeded after {i + 1} iterations"


@pytest.mark.skipif(sys.platform.lower() == "win32", reason="requires Unix or macOS")
@pytest.mark.skipif(machine().find("x86_64") == -1, reason="run only on x86_64")
def test_nclx_profile_leaks():
    mem_limit = None
    im_path = Path("images/heif_other/cat.hif")
    heif_file = pillow_heif.open_heif(im_path, convert_hdr_to_8bit=False)
    for i in range(800):
        _nclx = pillow_heif.private.read_color_profile(heif_file[0]._handle)  # noqa
        _nclx = None  # noqa
        gc.collect()
        mem = _get_mem_usage()
        if i < 300:
            mem_limit = mem + 1
            continue
        assert mem <= mem_limit, f"memory usage limit exceeded after {i + 1} iterations"



import os
from io import BytesIO
from pathlib import Path

import helpers
import pytest

from pillow_heif import from_pillow, open_heif, options, register_heif_opener

os.chdir(os.path.dirname(os.path.abspath(__file__)))
register_heif_opener()

pytest.importorskip("numpy", reason="NumPy not installed")
if not helpers.hevc_enc():
    pytest.skip("No HEVC encoder.", allow_module_level=True)


def test_primary_convert_to():
    im = helpers.gradient_rgb()
    im_heif = from_pillow(im)
    im_heif.add_from_pillow(im)
    im_heif.convert_to("BGR;16")
    assert im_heif[0].mode == "BGR;16"
    assert im_heif[1].mode == "RGB"
    im_heif = from_pillow(im)
    im_heif[0].convert_to("RGB;16")
    assert im_heif.mode == "RGB;16"


def test_convert_to_same_mode():
    im_heif = open_heif(helpers.create_heif())
    assert im_heif.mode == "RGB"
    im_heif.convert_to(im_heif.mode)
    assert im_heif.mode == "RGB"
    # Image should not be loaded, when target mode equal to current.
    assert not getattr(im_heif[0], "_img_data")


@pytest.mark.parametrize("enc_bits", (10, 12))
@pytest.mark.parametrize("mode", ("RGB;16", "BGR;16"))
def test_rgb8_to_16_bit_color_mode(mode, enc_bits):
    try:
        options.SAVE_HDR_TO_12_BIT = True if enc_bits == 12 else False
        png_pillow = helpers.gradient_rgb()
        heif_file = from_pillow(png_pillow)
        assert heif_file.bit_depth == 8
        heif_file.convert_to(mode)
        out_heic = BytesIO()
        heif_file.save(out_heic, quality=-1)
        assert heif_file.bit_depth == 16
        assert not heif_file.has_alpha
        heif_file = open_heif(out_heic, convert_hdr_to_8bit=False)
        assert heif_file.bit_depth == enc_bits
        assert not heif_file.has_alpha
        helpers.compare_hashes([png_pillow, out_heic], hash_size=8)
    finally:
        options.SAVE_HDR_TO_12_BIT = False


@pytest.mark.parametrize("enc_bits", (10, 12))
@pytest.mark.parametrize("mode", ("RGBA;16", "BGRA;16"))
def test_rgba8_to_16_bit_color_mode(mode, enc_bits):
    try:
        options.SAVE_HDR_TO_12_BIT = True if enc_bits == 12 else False
        png_pillow = helpers.gradient_rgba()
        heif_file = from_pillow(png_pillow)
        assert heif_file.bit_depth == 8
        heif_file.convert_to(mode)
        out_heic = BytesIO()
        heif_file.save(out_heic, quality=-1)
        assert heif_file.bit_depth == 16
        assert heif_file.has_alpha
        heif_file = open_heif(out_heic, convert_hdr_to_8bit=False)
        assert heif_file.bit_depth == enc_bits
        assert heif_file.has_alpha
        helpers.compare_hashes([png_pillow, out_heic], hash_size=8)
    finally:
        options.SAVE_HDR_TO_12_BIT = False


@pytest.mark.parametrize("img, bit", (("images/heif/RGB_10.heif", 10), ("images/heif/RGB_12.heif", 12)))
@pytest.mark.parametrize("mode", ("RGB;16", "BGR;16"))
def test_rgb_hdr_to_16bit_color_mode(img, mode, bit):
    heif_file = open_heif(Path(img), convert_hdr_to_8bit=False)
    assert heif_file.bit_depth == bit
    heif_file.convert_to(mode)
    out_heic = BytesIO()
    heif_file.save(out_heic, quality=-1)
    assert heif_file.bit_depth == 16
    assert not heif_file.has_alpha
    heif_file = open_heif(out_heic, convert_hdr_to_8bit=False)
    assert heif_file.bit_depth == 10
    assert not heif_file.has_alpha
    helpers.compare_hashes([Path(img), out_heic], hash_size=8)


@pytest.mark.parametrize("img, bit", (("images/heif/RGBA_10.heif", 10), ("images/heif/RGBA_12.heif", 12)))
@pytest.mark.parametrize("mode", ("RGBA;16", "BGRA;16"))
def test_rgba_hdr_to_16bit_color_mode(img, mode, bit):
    heif_file = open_heif(Path(img), convert_hdr_to_8bit=False)
    assert heif_file.bit_depth == bit
    heif_file.convert_to(mode)
    out_heic = BytesIO()
    heif_file.save(out_heic, quality=-1)
    assert heif_file.bit_depth == 16
    assert heif_file.has_alpha
    heif_file = open_heif(out_heic, convert_hdr_to_8bit=False)
    assert heif_file.bit_depth == 10
    assert heif_file.has_alpha
    helpers.compare_hashes([Path(img), out_heic], hash_size=8, max_difference=1)


def test_rgba16_to_rgba_color_mode():
    heif_file_orig = open_heif("images/heif/RGBA_10.heif", convert_hdr_to_8bit=True)
    heif_file = open_heif("images/heif/RGBA_10.heif", convert_hdr_to_8bit=False)
    heif_file.convert_to("RGBA;16")
    heif_file.convert_to("RGBA")
    helpers.assert_image_similar(heif_file.to_pillow(), heif_file_orig.to_pillow())



# With OpenCV we test BGR, BGRA, BGR;15 and BGRA;16 modes

import os
from io import BytesIO
from pathlib import Path

import pytest
from helpers import compare_hashes, gradient_rgb_bytes, gradient_rgba_bytes, hevc_enc
from PIL import Image

from pillow_heif import HeifImagePlugin  # noqa
from pillow_heif import from_bytes, open_heif, options

np = pytest.importorskip("numpy", reason="NumPy not installed")
cv2 = pytest.importorskip("cv2", reason="OpenCV not installed")

os.chdir(os.path.dirname(os.path.abspath(__file__)))


@pytest.mark.skipif(not hevc_enc(), reason="Requires HEVC encoder.")
@pytest.mark.parametrize("enc_bits", (10, 12))
def test_save_bgr_16bit_to_10_12_bit(enc_bits):
    try:
        options.SAVE_HDR_TO_12_BIT = True if enc_bits == 12 else False
        image_path = "images/non_heif/RGB_16.png"
        cv_img = cv2.imread(image_path, cv2.IMREAD_UNCHANGED)
        assert cv_img.shape[2] == 3  # 3 channels(BGR)
        heif_file = from_bytes(mode="BGR;16", size=(cv_img.shape[1], cv_img.shape[0]), data=bytes(cv_img))
        out_heic = BytesIO()
        heif_file.save(out_heic, quality=-1)
        heif_file = open_heif(out_heic, convert_hdr_to_8bit=False)
        assert heif_file.bit_depth == enc_bits
        png_pillow = Image.open(Path(image_path))
        heif_pillow = Image.open(out_heic)
        compare_hashes([png_pillow, heif_pillow], hash_size=8)
    finally:
        options.SAVE_HDR_TO_12_BIT = False


@pytest.mark.skipif(not hevc_enc(), reason="Requires HEVC encoder.")
@pytest.mark.parametrize("enc_bits", (10, 12))
def test_save_bgra_16bit_to_10_12_bit(enc_bits):
    try:
        options.SAVE_HDR_TO_12_BIT = True if enc_bits == 12 else False
        image_path = "images/non_heif/RGBA_16.png"
        cv_img = cv2.imread(image_path, cv2.IMREAD_UNCHANGED)
        assert cv_img.shape[2] == 4  # 4 channels(BGRA)
        heif_file = from_bytes(mode="BGRA;16", size=(cv_img.shape[1], cv_img.shape[0]), data=bytes(cv_img))
        out_heic = BytesIO()
        heif_file.save(out_heic, quality=-1)
        heif_file = open_heif(out_heic, convert_hdr_to_8bit=False)
        assert heif_file.bit_depth == enc_bits
        png_pillow = Image.open(Path(image_path))
        heif_pillow = Image.open(out_heic)
        compare_hashes([png_pillow, heif_pillow], hash_size=8, max_difference=1)
    finally:
        options.SAVE_HDR_TO_12_BIT = False


@pytest.mark.skipif(not hevc_enc(), reason="Requires HEVC encoder.")
def test_save_bgr_8bit_color_mode():
    rgb8_buf = gradient_rgb_bytes("PNG")
    cv_img = cv2.imdecode(np.asarray(rgb8_buf), cv2.IMREAD_UNCHANGED)
    assert cv_img.shape[2] == 3  # 3 channels(BGR)
    heif_file = from_bytes(mode="BGR", size=(cv_img.shape[1], cv_img.shape[0]), data=bytes(cv_img))
    out_heic = BytesIO()
    heif_file.save(out_heic, quality=-1)
    assert open_heif(out_heic, convert_hdr_to_8bit=False).bit_depth == 8
    png_pillow = Image.open(BytesIO(rgb8_buf))
    heif_pillow = Image.open(out_heic)
    compare_hashes([png_pillow, heif_pillow], hash_size=8)


@pytest.mark.skipif(not hevc_enc(), reason="Requires HEVC encoder.")
def test_save_bgra_8bit_color_mode():
    rgba8_buf = gradient_rgba_bytes("PNG")
    cv_img = cv2.imdecode(np.asarray(rgba8_buf), cv2.IMREAD_UNCHANGED)
    assert cv_img.shape[2] == 4  # 4 channels(BGRA)
    heif_file = from_bytes(mode="BGRA", size=(cv_img.shape[1], cv_img.shape[0]), data=bytes(cv_img))
    out_heic = BytesIO()
    heif_file.save(out_heic, quality=-1)
    assert open_heif(out_heic, convert_hdr_to_8bit=False).bit_depth == 8
    png_pillow = Image.open(BytesIO(rgba8_buf))
    heif_pillow = Image.open(out_heic)
    compare_hashes([png_pillow, heif_pillow], hash_size=8)


@pytest.mark.parametrize(
    "img_path",
    (
        "images/heif/RGB_10.heif",
        "images/heif/RGBA_10.heif",
        "images/heif/RGB_12.heif",
        "images/heif/RGBA_12.heif",
    ),
)
def test_read_10_12_bit(img_path):
    image_path = Path(img_path)
    heif_file = open_heif(image_path, convert_hdr_to_8bit=False)
    heif_file.convert_to("BGRA;16" if heif_file.has_alpha else "BGR;16")
    np_array = np.asarray(heif_file)
    img_encode = cv2.imencode(".png", np_array)[1]
    compare_hashes([BytesIO(img_encode), image_path], hash_size=8)



#!/usr/bin/env python
"""Script to build wheel"""
from setuptools import setup
from wheel.bdist_wheel import bdist_wheel


class WheelsABI3(bdist_wheel):
    def get_tag(self):
        python, abi, plat = super().get_tag()
        if python.startswith("cp"):
            python = "cp37"
            abi = "abi3"
            if plat.startswith("macosx") and plat.find("x86_64") == -1:
                python = "cp38"
        return python, abi, plat


def get_version():
    version_file = "pillow_heif/_version.py"
    with open(version_file, encoding="utf-8") as f:
        exec(compile(f.read(), version_file, "exec"))  # pylint: disable=exec-used
    return locals()["__version__"]


setup(
    version=get_version(),
    cffi_modules=["libheif/build.py:ffi"],
    cmdclass={"bdist_wheel": WheelsABI3},
)



from io import BytesIO

import helpers
import pytest
from packaging.version import parse as parse_version
from PIL import Image
from PIL import __version__ as pil_version
from PIL import features

import pillow_heif

pillow_heif.register_avif_opener()
pillow_heif.register_heif_opener()


@pytest.mark.skipif(not features.check("webp"), reason="Requires WEBP support.")
@pytest.mark.skipif(not helpers.aom_enc(), reason="Requires AVIF encoder.")
@pytest.mark.skipif(not helpers.hevc_enc(), reason="Requires HEVC encoder.")
@pytest.mark.skipif(parse_version(pil_version) < parse_version("9.2.0"), reason="Requires Pillow >= 9.2")
@pytest.mark.parametrize("save_format", ("HEIF", "AVIF"))
@pytest.mark.parametrize(
    "im_format",
    (
        "PNG",
        "JPEG",
        "TIFF",
        "WEBP",
    ),
)
def test_exif_from_pillow(im_format, save_format):
    def pil_image_with_exif():
        _exif = Image.Exif()
        _exif[0x010E] = exif_desc_value
        _ = BytesIO()
        Image.new("RGB", (10, 10), 0).save(_, format=im_format, exif=_exif)
        return _

    exif_desc_value = "this is a desc"
    im = Image.open(pil_image_with_exif())
    exif = im.getexif()  # noqa
    assert exif[0x010E] == exif_desc_value
    out_im_heif = BytesIO()
    im.save(out_im_heif, format=save_format)  # saving to HEIF(AVIF)
    im_heif = Image.open(out_im_heif)
    assert im_heif.info["exif"]
    assert isinstance(im_heif.info["exif"], bytes)
    exif = im_heif.getexif()  # noqa
    assert exif[0x010E] == exif_desc_value
    out_im = BytesIO()
    im_heif.save(out_im, format=im_format, exif=im_heif.getexif())  # saving back to original format
    im = Image.open(out_im)
    exif = im.getexif()  # noqa
    assert exif[0x010E] == exif_desc_value


@pytest.mark.skipif(not helpers.hevc_enc(), reason="Requires HEVC encoder.")
def test_pillow_exif_add_remove():
    exif_desc_value = "this is a desc"
    exif = Image.Exif()
    exif[0x010E] = exif_desc_value
    out_heif_no_exif = BytesIO()
    out_heif = BytesIO()
    im = Image.new("RGB", (15, 15), 0)
    # filling `exif` during save.
    im.save(out_heif, format="HEIF", exif=exif.tobytes())
    assert "exif" not in im.info or im.info["exif"] is None
    im_heif = Image.open(out_heif)
    assert im_heif.info["exif"]
    # filling `info["exif"]` before save.
    im.info["exif"] = exif.tobytes()
    im.save(out_heif, format="HEIF")
    im_heif = Image.open(out_heif)
    assert im_heif.info["exif"]
    # setting `exif` to `None` during save.
    im_heif.save(out_heif_no_exif, format="HEIF", exif=None)
    assert im_heif.info["exif"]
    im_heif_no_exif = Image.open(out_heif_no_exif)
    assert "exif" not in im_heif_no_exif.info or im_heif_no_exif.info["exif"] is None
    # filling `info["exif"]` with `None` before save.
    im_heif.info["exif"] = None
    im_heif.save(out_heif_no_exif, format="HEIF")
    im_heif_no_exif = Image.open(out_heif_no_exif)
    assert "exif" not in im_heif_no_exif.info or im_heif_no_exif.info["exif"] is None
    # removing `info["exif"]` before save.
    im_heif.info.pop("exif")
    im_heif.save(out_heif_no_exif, format="HEIF")
    im_heif_no_exif = Image.open(out_heif_no_exif)
    assert "exif" not in im_heif_no_exif.info or im_heif_no_exif.info["exif"] is None


@pytest.mark.skipif(not helpers.hevc_enc(), reason="Requires HEVC encoder.")
def test_heif_exif_add_remove():
    exif_desc_value = "this is a desc"
    exif = Image.Exif()
    exif[0x010E] = exif_desc_value
    out_heif_no_exif = BytesIO()
    out_heif = BytesIO()
    # test filling `exif` during save.
    im_heif = pillow_heif.from_pillow(Image.new("RGB", (15, 15), 0))
    im_heif.save(out_heif, exif=exif.tobytes())
    assert "exif" not in im_heif.info or im_heif.info["exif"] is None
    im_heif = pillow_heif.open_heif(out_heif)
    assert im_heif.info["exif"]
    # test filling `info["exif"]` before save.
    im_heif = pillow_heif.from_pillow(Image.new("RGB", (15, 15), 0))
    im_heif.info["exif"] = exif.tobytes()
    im_heif.save(out_heif)
    im_heif = pillow_heif.open_heif(out_heif)
    assert im_heif.info["exif"]
    # setting `exif` to `None` during save.
    im_heif.save(out_heif_no_exif, exif=None)
    assert im_heif.info["exif"]
    im_heif_no_exif = pillow_heif.open_heif(out_heif_no_exif)
    assert "exif" not in im_heif_no_exif.info or im_heif_no_exif.info["exif"] is None
    # filling `info["exif"]` with `None` before save.
    im_heif.info["exif"] = None
    im_heif.save(out_heif_no_exif)
    im_heif_no_exif = pillow_heif.open_heif(out_heif_no_exif)
    assert "exif" not in im_heif_no_exif.info or im_heif_no_exif.info["exif"] is None
    # removing `info["exif"]` before save.
    im_heif.info.pop("exif")
    im_heif.save(out_heif_no_exif)
    im_heif_no_exif = pillow_heif.open_heif(out_heif_no_exif)
    assert "exif" not in im_heif_no_exif.info or im_heif_no_exif.info["exif"] is None


@pytest.mark.skipif(not helpers.hevc_enc(), reason="Requires HEVC encoder.")
def test_heif_multi_frame_exif_add_remove():
    exif_desc_value = "this is a desc"
    exif = Image.Exif()
    exif[0x010E] = exif_desc_value
    out_heif_no_exif = BytesIO()
    out_heif = BytesIO()
    # test filling `exif` during save.
    im_heif = pillow_heif.from_pillow(Image.new("RGB", (15, 15), 0))
    im_heif.add_from_pillow(Image.new("RGB", (13, 13), 0))
    im_heif.save(out_heif, exif=exif.tobytes(), primary_index=1)
    for i in range(2):
        assert "exif" not in im_heif[i].info or im_heif[i].info["exif"] is None
    im_heif = pillow_heif.open_heif(out_heif)
    assert im_heif[1].info["exif"] and not im_heif[0].info["exif"]
    # test filling `info["exif"]` before save.
    im_heif = pillow_heif.from_pillow(Image.new("RGB", (15, 15), 0))
    im_heif.add_from_pillow(Image.new("RGB", (13, 13), 0))
    im_heif[1].info["exif"] = exif.tobytes()
    im_heif.save(out_heif, primary_index=1)
    im_heif = pillow_heif.open_heif(out_heif)
    assert im_heif[1].info["exif"] and not im_heif[0].info["exif"]
    # setting `exif` to `None` during save.
    im_heif.save(out_heif_no_exif, exif=None)
    assert im_heif[1].info["exif"] and not im_heif[0].info["exif"]
    im_heif_no_exif = pillow_heif.open_heif(out_heif_no_exif)
    assert "exif" not in im_heif_no_exif[0].info or im_heif_no_exif[0].info["exif"] is None
    assert "exif" not in im_heif_no_exif[1].info or im_heif_no_exif[1].info["exif"] is None
    # filling `info["exif"]` with `None` before save.
    im_heif.info["exif"] = None
    im_heif.save(out_heif_no_exif)
    im_heif_no_exif = pillow_heif.open_heif(out_heif_no_exif)
    assert "exif" not in im_heif_no_exif[0].info or im_heif_no_exif[0].info["exif"] is None
    assert "exif" not in im_heif_no_exif[1].info or im_heif_no_exif[1].info["exif"] is None
    # removing `info["exif"]` before save.
    im_heif.info.pop("exif")
    im_heif.save(out_heif_no_exif)
    im_heif_no_exif = pillow_heif.open_heif(out_heif_no_exif)
    assert "exif" not in im_heif_no_exif[0].info or im_heif_no_exif[0].info["exif"] is None
    assert "exif" not in im_heif_no_exif[1].info or im_heif_no_exif[1].info["exif"] is None


@pytest.mark.skipif(not helpers.hevc_enc(), reason="Requires HEVC encoder.")
def test_corrupted_exif():
    exif_data = b"This_is_not_valid_EXIF_data"
    out_im = BytesIO()
    helpers.gradient_rgb().save(out_im, format="HEIF", exif=exif_data)
    im = Image.open(out_im)
    with pytest.raises(SyntaxError):
        im.getexif()
    assert im.info["exif"] == exif_data


@pytest.mark.skipif(not helpers.hevc_enc(), reason="Requires HEVC encoder.")
def test_exif_as_class():
    exif = Image.Exif()
    exif[258] = 8
    exif[40963] = 455
    exif[305] = "exif test"
    out_im = BytesIO()
    helpers.gradient_rgb().save(out_im, format="HEIF", exif=exif)
    im = Image.open(out_im)
    exif = im.getexif()
    assert exif[258] == 8
    assert exif[40963] == 455
    assert exif[305] == "exif test"



from io import BytesIO
from os import getenv
from pathlib import Path
from typing import Union

from PIL import Image, ImageMath, ImageOps

from pillow_heif import HeifFile, HeifImage, HeifThumbnail, add_thumbnails, libheif_info

try:
    import numpy as np
except ImportError:
    np = None


RELEASE_FULL_FLAG = getenv("PH_FULL_ACTION", "0") == "1"
RELEASE_LIGHT_FLAG = getenv("PH_LIGHT_ACTION", "0") == "1"


def assert_image_equal(a, b):
    assert a.mode == b.mode
    assert a.size == b.size
    assert a.tobytes() == b.tobytes()


def convert_to_comparable(a, b):
    new_a, new_b = a, b
    if a.mode == "P":
        new_a = Image.new("L", a.size)
        new_b = Image.new("L", b.size)
        new_a.putdata(a.getdata())
        new_b.putdata(b.getdata())
    elif a.mode == "I;16":
        new_a = a.convert("I")
        new_b = b.convert("I")
    return new_a, new_b


def assert_image_similar(a, b, epsilon=0.0):
    assert a.mode == b.mode
    assert a.size == b.size
    a, b = convert_to_comparable(a, b)
    diff = 0
    for ach, bch in zip(a.split(), b.split()):
        ch_diff = ImageMath.eval("abs(a - b)", a=ach, b=bch).convert("L")
        diff += sum(i * num for i, num in enumerate(ch_diff.histogram()))
    ave_diff = diff / (a.size[0] * a.size[1])
    assert epsilon >= ave_diff


def compare_hashes(pillow_images: list, hash_size=16, max_difference=0):
    if np is None:
        return
    image_hashes = []
    for pillow_image in pillow_images:
        if isinstance(pillow_image, (str, Path, BytesIO)):
            pillow_image = Image.open(pillow_image)
        pillow_image = ImageOps.exif_transpose(pillow_image)
        pillow_image = pillow_image.convert("L").resize((hash_size + 1, hash_size), Image.ANTIALIAS)
        pixels = np.asarray(pillow_image)  # noqa
        diff = pixels[:, 1:] > pixels[:, :-1]
        image_hash = diff.flatten()  # noqa
        for _ in range(len(image_hashes)):
            distance = np.count_nonzero(image_hash != image_hashes[_])
            assert distance <= max_difference, f"{distance} > {max_difference}"
        image_hashes.append(image_hash)


def compare_heif_files_fields(
    heif1: Union[HeifFile, HeifImage], heif2: Union[HeifFile, HeifImage], ignore=None, thumb_size_max_differ=2
):
    def compare_images_fields(image1: HeifImage, image2: HeifImage):
        assert image1.size == image2.size
        assert image1.mode == image2.mode
        if "original_bit_depth" not in ignore:
            assert image1.original_bit_depth == image2.original_bit_depth
        assert image1.bit_depth == image2.bit_depth
        if "stride" not in ignore:
            assert image1.stride == image2.stride
            assert len(image1.data) == len(image2.data)
        for i_thumb, thumbnail in enumerate(image1.thumbnails):
            with_difference = thumbnail.size[0] - image2.thumbnails[i_thumb].size[0]
            height_difference = thumbnail.size[1] - image2.thumbnails[i_thumb].size[1]
            assert with_difference + height_difference <= thumb_size_max_differ
            assert thumbnail.mode == image2.thumbnails[i_thumb].mode
            if "original_bit_depth" not in ignore:
                assert thumbnail.original_bit_depth == image2.thumbnails[i_thumb].original_bit_depth
            assert thumbnail.bit_depth == image2.thumbnails[i_thumb].bit_depth
        assert image1.info["exif"] == image2.info["exif"]
        assert image1.info["xmp"] == image2.info["xmp"]
        for block_i, block in enumerate(image1.info["metadata"]):
            assert block["data"] == image1.info["metadata"][block_i]["data"]
            assert block["content_type"] == image1.info["metadata"][block_i]["content_type"]
            assert block["type"] == image1.info["metadata"][block_i]["type"]

    if ignore is None:
        ignore = []
    if isinstance(heif1, HeifFile):
        for i, image in enumerate(heif1):
            compare_images_fields(image, heif2[i])
    else:
        compare_images_fields(heif1, heif2)


def compare_heif_to_pillow_fields(heif: Union[HeifFile, HeifImage, HeifThumbnail], pillow: Image):
    def compare_images_fields(heif_image: Union[HeifImage, HeifThumbnail], pillow_image: Image):
        assert heif_image.size == pillow_image.size
        assert heif_image.mode == pillow_image.mode
        for k in ("exif", "xmp", "metadata"):
            if heif_image.info.get(k, None):
                if isinstance(heif_image.info[k], (bool, int, float, str)):
                    assert heif_image.info[k] == pillow_image.info[k]
                else:
                    assert len(heif_image.info[k]) == len(pillow_image.info[k])
        assert heif_image.info.get("nclx_profile", None) == pillow_image.info.get("nclx_profile", None)
        for k in ("icc_profile", "icc_profile_type"):
            if heif_image.info.get(k, None):
                assert len(heif_image.info[k]) == len(pillow_image.info[k])

    if isinstance(heif, HeifFile):
        for i, image in enumerate(heif):
            pillow.seek(i)
            compare_images_fields(image, pillow)
    else:
        compare_images_fields(heif, pillow)


def create_heif(size: tuple = None, thumb_boxes: list = None, n_images=1, **kwargs) -> BytesIO:
    if size is None:
        size = (512, 512)
    if thumb_boxes is None:
        thumb_boxes = []
    im_heif = HeifFile()
    for i in range(n_images):
        im_heif.add_from_pillow(Image.effect_mandelbrot(size, (-3, -2.5, 2, 2.5), 100))
        size = (int(size[0] / 2), int(size[1] / 2))
        add_thumbnails(im_heif[i], thumb_boxes)
    heif_buf = BytesIO()
    im_heif.save(heif_buf, **kwargs)
    return heif_buf


def gradient_rgb():
    return Image.merge(
        "RGB",
        [
            Image.linear_gradient(mode="L"),
            Image.linear_gradient(mode="L").transpose(Image.ROTATE_90),
            Image.linear_gradient(mode="L").transpose(Image.ROTATE_180),
        ],
    )


def gradient_rgb_bytes(im_format: str) -> bytearray:
    _ = BytesIO()
    gradient_rgb().save(_, format=im_format)
    return bytearray(_.getbuffer().tobytes())


def gradient_rgba():
    return Image.merge(
        "RGBA",
        [
            Image.linear_gradient(mode="L"),
            Image.linear_gradient(mode="L").transpose(Image.ROTATE_90),
            Image.linear_gradient(mode="L").transpose(Image.ROTATE_180),
            Image.linear_gradient(mode="L").transpose(Image.ROTATE_270),
        ],
    )


def gradient_rgba_bytes(im_format: str) -> bytearray:
    _ = BytesIO()
    gradient_rgba().save(_, format=im_format)
    return bytearray(_.getbuffer().tobytes())


def gradient_la():
    return Image.merge(
        "LA",
        [
            Image.linear_gradient(mode="L"),
            Image.linear_gradient(mode="L").transpose(Image.ROTATE_90),
        ],
    )


def gradient_la_bytes(im_format: str) -> bytearray:
    _ = BytesIO()
    gradient_la().save(_, format=im_format)
    return bytearray(_.getbuffer().tobytes())


def gradient_p_bytes(im_format: str) -> bytearray:
    _ = BytesIO()
    im = Image.linear_gradient(mode="L")
    im = im.convert(mode="P")
    im.save(_, format=im_format)
    return bytearray(_.getbuffer().tobytes())


def gradient_pa():
    return Image.merge(
        "PA",
        [
            Image.linear_gradient(mode="L"),
            Image.linear_gradient(mode="L").transpose(Image.ROTATE_90),
        ],
    )


def gradient_pa_bytes(im_format: str) -> bytearray:
    _ = BytesIO()
    gradient_la().save(_, format=im_format)
    return bytearray(_.getbuffer().tobytes())


def hevc_enc() -> bool:
    if getenv("PH_TESTS_NO_HEVC_ENC", "0") != "0":
        return False
    return libheif_info()["encoders"]["HEVC"]


def aom_dec() -> bool:
    if getenv("PH_TESTS_NO_AVIF_DEC", "0") != "0":
        return False
    if libheif_info()["version"]["aom"] == "Rav1e encoder":
        return False
    return libheif_info()["decoders"]["AV1"]


def aom_enc() -> bool:
    if getenv("PH_TESTS_NO_AVIF_ENC", "0") != "0":
        return False
    if libheif_info()["version"]["aom"] == "Rav1e encoder":
        return False
    return libheif_info()["encoders"]["AV1"]
